from typing import Union, Tuple, Optional, Dict, List
import blessed
import threading
import time
import math
import traceback

from tegen.scene import Scene
from tegen.objects import Screen, Sprite, Object, Text
import tegen.pixel as pixel

class Game:
    """The entry point for the game.
    
    .. versionadded:: 0.0
    
    .. py:attribute:: game_on
       :type: bool
       
       Whether the game is running.
       
       .. versionadded:: 0.0
       
    .. py:attribute:: loop
       :type: threading.Thread
       
       The looping thread of the game.
       
       .. versionadded:: 0.0

    .. py:attribute:: keyboard_listener
       :type: threading.Thread

       The keyboard listening thread of the game.

       .. versionadded:: 0.0
       
    .. py:attribute:: objects
       :type: Dict[object, dict]
       
       A list of objects currently loaded.
       
       .. versionadded:: 0.0

    .. py:attribute:: screen
       :type: Screen

       The screen object of the game.

       .. versionadded:: 0.0

    .. py:attribute:: current_scene
       :type: Scene

       The current scene that the game is showing.

       .. versionadded:: 0.0

    .. py:attribute:: speeds
       :type: List[float]

       A list of milliseconds per frame.

       .. versionadded:: 0.0"""
    
    term = blessed.Terminal()

    def __init__(self):
        self.game_on = False
        self.loop: threading.Thread = None
        self.keyboard_listener: threading.Thread = None
        self.objects: Dict[tuple, Object] = {}
        self.screen: Screen = Screen(0, 0)
        self.current_scene: Scene = None
        self.speeds: List[float] = []

    def start(self, show_info: bool=True, info_wait: Union[int, float]=3):
        """Starts the game.

        .. versionadded:: 0.0

        :param bool show_info: Whether to show tegen and terminal info before the game starts
        :param info_wait: The amount of time for tegen and terminal info to show
        :type info_wait: int or float"""
        term = self.term
        print(term.height*"\n")
        print(term.home + term.clear, end='')
        if show_info:
            from tegen import __version__
            print(term.bold("tegen v"+__version__))
            print("number of colours: "+str(term.number_of_colors))
            print("terminal size (h,w): "+str((term.height, term.width)))
            time.sleep(info_wait)
        print(term.home + term.clear, end='')
        self.game_on = True
        self.loop = threading.Thread(target=_loop, args=(self,))
        self.loop.start()
        
    def end(self):
        """Ends the game.

        .. versionadded:: 0.0"""
        term = self.term
        for id_, obj in self.objects.items():
            obj.on_end(self)
        self.game_on = False
        print(term.home + term.clear + term.bright_yellow("Stopping..."), end='')
        time.sleep(0.5)
        print(term.home + term.clear, end='')

    def load_scene(self, scene: Scene, clear_objects: bool=True):
        """Loads a scene to the game.

        .. versionadded:: 0.0

        :param Scene scene: The scene to load
        :param bool clear_objects: Whether to clear all objects in the previous scene before loading the new scene"""
        for id_, obj in self.objects.items():
            obj.on_end(self)
        self.current_scene = scene
        if clear_objects: self.objects.clear()
        self.objects.update(scene.objects)
        for id_, obj in self.objects.items():
            obj.on_init(self)

    def call_event(self, event: str, *args, **kwargs):
        """Calls an event, running `on_<event name>` in all :py:class:`Object` s, if present.
        
        .. versionadded:: 0.0
        
        :param str event: The name of the event to call"""
        def empty():
            pass

        items = list(self.objects.items())[:]
        for id_, obj in items:
            getattr(obj, 'on_'+event, empty)(self, *args, **kwargs) # noqa

    def add_object(self, obj: Object, id_: str, x: float, y: float, override: bool = False):
        """Adds an :py:class:`Object` to the game.

        .. versionadded:: 0.0

        :param Object obj: The object to add
        :param str id_: The ID to give to the object
        :param float x: The global x coordinate of the anchor (local x=0)
        :param float y: The global y coordinate of the anchor (local y=0)
        :param bool override: Whether to override the existing object, if an object with the same ID exists
        :raises ValueError: if the ``obj`` is a :py:class:`Screen`
        :raises KeyError: if an object with the same ID exists and ``override`` is False"""
        if isinstance(obj, Screen):
            raise ValueError("Object added cannot be a screen, access the screen via `Game.screen`")
        if id_ in self.objects.keys() and not override:
            raise KeyError(f"Object '{id_}' already exists")
        obj.id = id_
        obj.x = x
        obj.y = y
        self.objects[id_] = obj

    def remove_object_by_id(self, id_: str, nonexist_error: bool = False):
        """Removes an :py:class:`Object` from the game by its ID.

        .. versionadded:: 0.0

        :param str id_: The ID of the object to remove
        :param bool nonexist_error: Whether to raise an error if an object does not exist in the game."""
        try:
            del self.objects[id_]
        except KeyError as e:
            if nonexist_error: raise e

    def remove_object_by_class(self, cls: type):
        """Removes :py:class:`Object` s from the game by their class type.

        .. versionadded:: 0.0

        :param type cls: The class, should be a subclass of :py:class:`Object`
        :raises TypeError: if the class is not a subclass of :py:class:`Object`"""
        if not issubclass(cls, Object):
            raise TypeError("Class is not subclass of Object")
        for id_, obj in self.objects.items():
            if isinstance(obj, cls):
                del self.objects[id_]

    def add_keyboard_listener(self):
        """Adds a keyboard listener, to fire events when a key is pressed.

        .. versionadded:: 0.0"""
        self.keyboard_listener = threading.Thread(target=_keyboard, args=(self,))
        self.keyboard_listener.start()
            
    def mspf(self) -> Union[Union[float, int], None]:
        """Gets the number of milliseconds per frame.
        
        .. versionadded:: 0.0

        :rtype: float or int or None"""
        if len(self.speeds) != 0: return sum(self.speeds)/len(self.speeds)
        return None

    def fps(self) -> Union[Union[float, int], None]:
        """Gets the number of frames per second.
        
        .. versionadded:: 0.0
        
        :rtype: float or int or None"""
        if len(self.speeds) != 0: avg_ms = sum(self.speeds)/len(self.speeds)
        else: return None
        if avg_ms == 0: return math.inf
        return round(1000 / avg_ms, 2)

    def get_displayed_pixel(self, x: int, y: int) -> Tuple[Optional[Tuple[int, int, int]], Optional[Tuple[int, int, int]], Optional[str]]:
        """Get the pixel at a certain global coordinate.
        
        .. versionadded:: 0.0
        
        :param int x: The global x coordinate of the pixel.
        :param int y: The global y coordinate of the pixel.
        :returns: A tuple of ``(back colour, fore colour, character)``
        :rtype: Tuple[Optional[Tuple[int, int, int]], Optional[Tuple[int, int, int]], Optional[str]]"""
        back, fore, char = (None,)*3
        values = list(self.objects.values())[:]
        for obj in values:
            lx, rx, ty, by = obj.edges()
            if x < lx or x > rx or y < ty or y > by: continue
            if issubclass(type(obj), Sprite):
                obj: Sprite # pacify linter
                pixel_info = obj.pixels[x-obj.x, y-obj.y]
                if 'back' in pixel_info.keys() and pixel_info['back'] is not None: back = pixel_info['back']
                if 'fore' in pixel_info.keys() and ['fore'] is not None: fore = pixel_info['fore']
                if 'char' in pixel_info.keys() and ['char'] is not None: char = pixel_info['char']
            elif issubclass(type(obj), Text):
                obj: Text # pacify linter
                pos = obj.get_char_positions()
                if (x-obj.x, y-obj.y) not in pos.keys(): continue
                if obj.back is not None: back = pixel._parse_colours(obj.back) # noqa
                if obj.fore is not None: fore = pixel._parse_colours(obj.fore) # noqa
                char = pos[x-obj.x, y-obj.y]
            else:
                continue
        return back, fore, char

    def handle_error(self):
        """Handles any error properly when the game is running.

        .. versionadded:: 0.0

        **Example:**

        .. code-block:: python

           try:
               game.start()
               ...
           except Exception as e:
               game.handle_error()
        """
        term = self.term
        self.game_on = False
        print(term.home + term.clear_eos + term.bright_red("An error has occured and the game will quit shortly.\n") + term.red(traceback.format_exc()))
        time.sleep(0.5)
        print(term.bright_red("Press any key to continue..."))
        with term.cbreak():
            term.inkey(timeout=15)
        

def _loop(game: Game):
    """:meta private:"""
    term = game.term
    try:
        while game.game_on:
            loop_start = time.time()
            for obj in game.objects.values():
                obj.pre_update(game)
            for obj in game.objects.values():
                obj.update(game)
            for obj in game.objects.values():
                obj.post_update(game)

            out = ""
            lx, rx, ty, by = game.screen.edges()
            for y in range(ty, by+1):
                for x in range(lx, rx+1):
                    back, fore, char = game.get_displayed_pixel(x, y)
                    back_style = (lambda o: o) if back is None else term.on_color_rgb(*back)
                    fore_style = (lambda o: o) if fore is None else term.color_rgb(*fore)
                    char = " " if char is None else char
                    out += back_style(fore_style(char))
            print(term.home + out + term.clear_eos, end="", flush=True)

            #print(term.home + str(game.fps()) + term.clear_eol, flush=True)
            game.speeds.append(1000*(time.time()-loop_start))
            if len(game.speeds) > 100: game.speeds.pop(0)
    except Exception:
        game.handle_error()

def _keyboard(game: Game):
    """:meta private:"""
    term = game.term
    try:
        while game.game_on:
            with term.cbreak():
                key = term.inkey(timeout=1)
                if key: game.call_event("keyboard_press", key)
    except Exception:
        game.handle_error()


from typing import Dict

from tegen.objects import Screen, Object

class Scene:
    """A game scene.
    
    .. versionadded:: 0.0
    
    .. py:attribute:: objects
       :type: Dict[str, Object]
       
       A dict of objects, in the form of ``{id: object}``"""

    def __init__(self):
        self.objects: Dict[str, Object] = {}

    def add_object(self, obj: Object, id_: str, x: float, y: float, override: bool=False):
        """Adds an :py:class:`Object` to the scene.
        
        .. versionadded:: 0.0
        
        :param Object obj: The object to add
        :param str id_: The ID to give to the object
        :param float x: The global x coordinate of the anchor (local x=0)
        :param float y: The global y coordinate of the anchor (local y=0)
        :param bool override: Whether to override the existing object, if an object with the same ID exists
        :raises ValueError: if the ``obj`` is a :py:class:`Screen`
        :raises KeyError: if an object with the same ID exists and ``override`` is False"""
        if isinstance(obj, Screen):
            raise ValueError("Object added cannot be a screen, access the screen via `Game.screen`")
        if id_ in self.objects.keys() and not override:
            raise KeyError(f"Object '{id_}' already exists")
        obj.id = id_
        obj.x = x
        obj.y = y
        self.objects[id_] = obj

    def remove_object_by_id(self, id_: str, nonexist_error: bool=False):
        """Removes an :py:class:`Object` from the scene by its ID.

        .. versionadded:: 0.0

        :param str id_: The ID of the object to remove
        :param bool nonexist_error: Whether to raise an error if an object does not exist in the scene."""
        try:
            del self.objects[id_]
        except KeyError as e:
            if nonexist_error: raise e

    def remove_object_by_class(self, cls: type):
        """Removes :py:class:`Object` s from the scene by their class type.
        
        .. versionadded:: 0.0
        
        :param type cls: The class, should be a subclass of :py:class:`Object`
        :raises TypeError: if the class is not a subclass of :py:class:`Object`"""
        if not issubclass(cls, Object):
            raise TypeError("Class is not subclass of Object")
        for id_, obj in self.objects.items():
            if isinstance(obj, cls):
                del self.objects[id_]

        



# Example code for tegen: Tic Tac Toe, by 7d
# Enjoy my shoddy code :)
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))
import tegen
import blessed
from blessed.keyboard import Keystroke

game = tegen.Game()
scene = tegen.Scene()
term = blessed.Terminal()

class Lines(tegen.objects.Sprite):
    pixels = tegen.pixel.from_2d_array(char=['     █     █     ',
                                             '     █     █     ',
                                             '     █     █     ',
                                             '█████████████████',
                                             '     █     █     ',
                                             '     █     █     ',
                                             '     █     █     ',
                                             '█████████████████',
                                             '     █     █     ',
                                             '     █     █     ',
                                             '     █     █     '])

class Label(tegen.objects.Text):
    fore = 0xffa500

class XPiece(tegen.objects.Sprite):
    pixels = tegen.pixel.from_2d_array(char=['█   █',
                                             ' ███ ',
                                             '█   █'],
                                       fore=[[0xff3300]*5]*3)

class OPiece(tegen.objects.Sprite):
    pixels = tegen.pixel.from_2d_array(char=[' ███ ',
                                             '█   █',
                                             ' ███ '],
                                       fore=[[0x00ccff]*5]*3)


class Notif(tegen.objects.Text):
    turn = 1
    won = False
    x_locations = []
    o_locations = []

    def on_keyboard_press(self, g: tegen.Game, key: Keystroke):
        if self.won or key == "q":
            g.end()
            return
        if key not in [str(k) for k in range(1, 10)]: return
        piece_x = g.objects["label" + str(key)].x-2
        piece_y = g.objects["label" + str(key)].y-1
        piece = XPiece if self.turn == 1 else OPiece
        if "piece" + str(key) in g.objects.keys(): return
        game.add_object(piece(), "piece" + str(key), piece_x, piece_y)

        locations = self.x_locations if piece == XPiece else self.o_locations
        locations.append(int(key))
        if all(c in locations for c in [1, 2, 3]) or \
           all(c in locations for c in [4, 5, 6]) or \
           all(c in locations for c in [7, 8, 9]) or \
           all(c in locations for c in [1, 4, 7]) or \
           all(c in locations for c in [2, 5, 8]) or \
           all(c in locations for c in [3, 6, 9]) or \
           all(c in locations for c in [1, 5, 9]) or \
           all(c in locations for c in [3, 5, 7]):
            self.won = True
            self.text = "Player " + str(self.turn) + " wins! Press any key to exit"
        elif sorted(self.x_locations+self.o_locations) == list(range(1, 10)):
            self.fore = 0xffcc00
            self.won = True
            self.text = "Drawed. Press any key to exit"
        else:
            self.turn = 2 if self.turn == 1 else 1
            self.fore = 0x00ccff if self.turn == 2 else 0xff3300
            self.text = "Player " + str(self.turn) + "'s turn"
        with term.cbreak():
            while term.inkey(timeout=0.1) != "":
                pass


scene.add_object(Lines(), "lines", 0, 0)
scene.add_object(Notif("Player 1's turn", fore=0xff3300), "notif", 0, 12)
count = 1
for y in [1, 5, 9]:
    for x in [2, 8, 14]:
        scene.add_object(Label(str(count)), "label"+str(count), x, y)
        count += 1

try:
    game.start(info_wait=1)
    game.add_keyboard_listener()
    game.load_scene(scene)
except Exception:
    game.handle_error()


from typing import List, Tuple, Dict, Optional
import blessed
from blessed.keyboard import Keystroke

import tegen.pixel as pixel


class Object:
    """The base class of all objects.
    
    .. versionadded:: 0.0
    
    .. py:attribute:: x
       :type: int
    
       The global x coordinate of the object, set when added to a scene
       
       .. versionadded:: 0.0
       
    .. py:attribute:: y
       :type: int
    
       The global y coordinate of the object, set when added to a scene
       
       .. versionadded:: 0.0
       
    .. py:attribute:: id
       :type: str
    
       The ID of the object, set when added to a scene
       
       .. versionadded:: 0.0"""

    def __init__(self):
        self.x: int = None
        self.y: int = None
        self.id: str = None

    def edges(self):
        pass

    def on_init(self, g):
        """This method is to be overridden when extended.
        Called on scene load.

        .. versionadded:: 0.0

        :param Game g: The game object"""

    def on_end(self, g):
        """This method is to be overridden when extended.
        Called on scene unload.

        .. versionadded:: 0.0

        :param Game g: The game object"""

    def pre_update(self, g):
        """This method is to be overridden when extended.
        Called every tick of the game loop, before :py:meth:`update`.

        .. versionadded:: 0.0

        :param Game g: The game object"""

    def update(self, g):
        """This method is to be overridden when extended.
        Called every tick of the game loop.

        .. versionadded:: 0.0

        :param Game g: The game object"""

    def post_update(self, g):
        """This method is to be overridden when extended.
        Called every tick of the game loop, after :py:meth:`update`.

        .. versionadded:: 0.0

        :param Game g: The game object"""

    def on_keyboard_press(self, g, key: Keystroke):
        """This method is to be overridden when extended.
        Called on a key press.

        .. versionadded:: 0.0

        :param Game g: The game object
        :param Keystroke key: The key pressed"""


class Screen(Object):
    """Inherited from :py:class:`Object`. Represents the screen.
    
    .. versionadded:: 0.0
    
    :param int x: The game x coordinate of the topleft corner
    :param int y: The game y coordinate of the topleft corner"""

    def __init__(self, x: int, y: int):
        self.x = x
        self.y = y

    def corners(self) -> List[Tuple[int, int]]:
        """Returns the global coordinates of the four corners of the screen.
        
        .. versionadded:: 0.0
        
        :returns: A list of coordinates, in the form ``[tl, tr, bl, br]``
        :rtype: List[Tuple[int, int]]"""
        term = blessed.Terminal()
        tl = self.x, self.y
        tr = self.x+term.width-1, self.y
        bl = self.x, self.y+term.height-1
        br = self.x+term.width-1, self.y+term.height-1
        return [tl, tr, bl, br]

    def edges(self) -> Tuple[int, int, int, int]:
        """Returns the global x coordinate of the leftmost and rightmost columns,
        and the global y coordinate of the topmost and bottommost rows of the screen.

        .. versionadded:: 0.0

        :returns: A tuple of values, in the form ``[lx, rx, ty, by]``
        :rtype: Tuple[int, int, int, int]"""
        term = blessed.Terminal()
        lx = self.x
        rx = self.x+term.width-1
        ty = self.y
        by = self.y+term.height-1
        return lx, rx, ty, by


class Sprite(Object):
    """Inherited from :py:class:`Object`. Represents a sprite.

    .. versionadded:: 0.0"""
    pixels = pixel.from_2d_array(fore=[['f00', 'aaa', 'f00'],
                                       ['aaa', 'f00', 'aaa'],
                                       ['f00', 'aaa', 'f00']],
                                 char=['███',
                                       '███',
                                       '███'])

    def edges(self) -> Tuple[int, int, int, int]:
        """Returns the global x coordinate of the leftmost and rightmost columns,
        and the global y coordinate of the topmost and bottommost rows of the screen.

        .. versionadded:: 0.0

        :returns: A tuple of values, in the form ``[lx, rx, ty, by]``
        :rtype: Tuple[int, int, int, int]"""
        lx, ty = (float("inf"),)*2
        rx, by = (0,)*2
        for local_x, local_y in list(self.pixels.keys()):
            if local_x > rx: rx = local_x
            if local_x < lx: lx = local_x
            if local_y > by: by = local_y
            if local_y < ty: ty = local_y
        
        return self.x+lx, self.x+rx, self.y+ty, self.y+by

    def local_move(self, x: int, y: int):
        """Move the sprite's local coordinates.

        .. versionadded:: 0.0

        :param x: The local x value to move the coordinates by
        :param y: The local y value to move the coordinates by
        """
        new_pixels = {}
        for coords, pixel_dict in self.pixels.items():
            new_coords = (coords[0]+x, coords[1]+y)
            new_pixels[new_coords] = pixel_dict

class Text(Object):
    """Represents some text on a screen.

    .. versionadded:: 0.0

    :param str text: The text to start with
    :param back: The background colour of the text
    :type back: Tuple[int, int, int]
    :param fore: The foreground colour of the text
    :type fore: Tuple[int, int, int]
    
    .. py:attribute:: anchor
       :type: str
    
       The anchor of text

       .. versionadded:: 0.0
       
    .. py:attribute:: text
       :type: str
       
       The content of the text

       .. versionadded:: 0.0

    .. py:attribute:: back
       :type: Tuple[int, int, int]

       The background colour of the text

       .. versionadded:: 0.0

    .. py:attribute:: fore
       :type: Tuple[int, int, int]

       The foreground colour of the text

       .. versionadded:: 0.0"""

    anchor = 'tl'
    back: Tuple[int, int, int] = None
    fore: Tuple[int, int, int] = None

    def __init__(self, text: str, back: Optional[pixel.Colour]=None, fore: Optional[pixel.Colour]=None):
        super().__init__()
        self.text = text
        if back is not None: self.back = pixel._parse_colours(back) # noqa
        if fore is not None: self.fore = pixel._parse_colours(fore) # noqa

    def edges(self) -> Tuple[int, int, int, int]:
        """Returns the global x coordinate of the leftmost and rightmost columns,
        and the global y coordinate of the topmost and bottommost rows of the screen.

        .. versionadded:: 0.0

        :returns: A tuple of values, in the form ``[lx, rx, ty, by]``
        :rtype: Tuple[int, int, int, int]"""
        w = max([len(l) for l in self.text.split("\n")])
        h = self.text.count('\n')+1
        return self.x, self.x+w-1, self.y, self.y+h-1

    def get_char_positions(self) -> Dict[tuple, str]:
        """Get the positions of each character relative to the anchor.

        .. versionadded:: 0.0

        :returns: A dict in the form ``{(local x, local y): char}``
        :rtype: Dict[tuple, str]
        :raises ValueError: if ``anchor`` is not one of ``tr``, ``tl``, ``br``, ``bl``, ``center``"""
        text = self.text
        w = max([len(l) for l in text.split("\n")])
        h = text.count('\n')
        if self.anchor == "center":
            ox = round(w / 2)
            oy = round(h / 2)
        else:
            if self.anchor not in ['tr', 'tl', 'br', 'bl']:
                raise ValueError("'anchor' is not one of 'center', 'tr', 'tl', 'br', 'bl'")
            ox = 0 if self.anchor[1] == 'l' else w - 0
            oy = 0 if self.anchor[0] == 't' else h - 0
        result = {}
        for line_num, line in enumerate(text.split('\n')):
            for char_num, char in enumerate(line):
                result[char_num-ox, line_num-oy] = char
        return result


from typing import List, Optional, Union, Tuple, Dict
import re
from PIL import Image

PixelMap = Dict[tuple, Dict[str, Union[str, Tuple[int, int, int]]]]
Colour = Union[Union[int, str], Union[tuple, list]]

def _parse_colours(colour: Optional[Colour]) -> Optional[Tuple[int, int, int]]:
    """:meta private:"""
    if colour is None: return None
    if isinstance(colour, (list, tuple)):
        if len(colour) != 3:
            raise ValueError(f"Colour sequence must have 3 values (Got {colour})")
        return tuple(colour)
    if isinstance(colour, str):
        m = re.search(r"^#?([0-9a-f]{6}|[0-9a-f]{3})$", colour)
        if m is None:
            raise ValueError(f"Invalid hex code")
        hexc = m.group(1)
        if len(hexc) == 3: hexc = 2*hexc[0]+2*hexc[1]+2*hexc[2]
        return int(hexc[0:2], base=16), int(hexc[2:4], base=16), int(hexc[4:6], base=16)
    elif isinstance(colour, int):
        hexc = hex(colour)[2:]
        while len(hexc) < 6: hexc = "0"+hexc
        return int(hexc[0:2], base=16), int(hexc[2:4], base=16), int(hexc[4:6], base=16)

def _find_origin(order: Tuple[int, int], anchor: str):
    """:meta private:"""
    lx, ty = (float("inf"),) * 2
    rx, by = (0,) * 2
    for y in range(order[1]):
        for x in range(order[0]):
            if x > rx: rx = x
            if x < lx: lx = x
            if y > by: by = y
            if y < ty: ty = y
    if anchor == "center":
        ox = round((lx + rx) / 2)
        oy = round((ty + by) / 2)
    else:
        if anchor not in ['tr', 'tl', 'br', 'bl']:
            raise ValueError("'anchor' is not one of 'center', 'tr', 'tl', 'br', 'bl'")
        ox = lx if anchor[1] == 'l' else rx
        oy = ty if anchor[0] == 't' else by
    return ox, oy

def from_2d_array(back: Optional[List[List[str]]]=None, fore: Optional[List[List[str]]]=None,
                  char: Optional[List[str]]=None, anchor: str='tl') -> PixelMap:
    """Generates a map of pixels from 2d arrays.

    .. versionadded:: 0.0

    :param List[List[str]] back: A list of lists of colours as the background
    :param List[List[str]] fore: A list of lists of colours as the foreground
    :param List[str] char: A list of strings as rows as the characters
    :param str anchor: The corner to set the local coordinate as ``(0, 0)``, choose from ``tr``, ``tl``, ``br``, ``bl``, ``center``
    :rtype: PixelMap
    :raises ValueError: if ``back``, ``fore``, or ``char`` has no columns, or is a list of empty rows
    :raises ValueError: if ``back``, ``fore``, or ``char`` has inconsistent lengths of rows
    :raises ValueError: if ``back``, ``fore``, or ``char`` have inconsistent array sizes"""

    #check for equal order of lists
    order = (0, 0)
    prev = ""
    if char is not None: char = [list(s) for s in char]
    for name, a in [('back', back), ('fore', fore), ('char', char)]:
        if a is None: continue
        if len(a) == 0:
            raise ValueError(f"Parameter {name} has no columns, perhaps change it to 'None'?")
        orders = [len(l) for l in a]
        if any(x != orders[0] for x in orders):
            raise ValueError(f"Inconsistent length of rows for parameter {name}")
        if orders[0] == 0:
            raise ValueError(f"Parameter {name} is made of columns of no elements, perhaps change it to 'None'?")
        this_order = (len(a), orders[0])
        if order == (0, 0):
            order = this_order
            prev = name
        elif order != this_order:
            raise ValueError(f"Inconsistent sizes of arrays: {prev} is of size {order} but {name} is of size {this_order}")
        else:
            order = this_order
            prev = name

    # find anchor
    ox, oy = _find_origin(order, anchor)

    result = {}
    for name, a in [('back', back), ('fore', fore), ('char', char)]:
        if a is None: continue
        for y, yv in enumerate(a):
            for x, xv in enumerate(yv):
                if (x-ox, y-oy) not in result.keys(): result[x-ox, y-oy] = {}
                v = xv if name == 'char' else _parse_colours(xv)
                if isinstance(v, str) and v.strip() == '': v = None
                result[x-ox, y-oy][name] = v
    return result

def from_image(fp: str, anchor: str='tl', layer: str='fore', char: str='█') -> PixelMap:
    """Generates a map of pixels from an image. Each pixel in the image represents one character in the terminal.

    :param str fp: The file path of the image
    :param str anchor: The corner to set the local coordinate as ``(0, 0)``, choose from ``tr``, ``tl``, ``br``, ``bl``, ``center``
    :param str layer: The layer to write the pixels to, choose from ``back``, ``fore``
    :param str char: The character to serve as the single pixel
    :rtype: PixelMap
    :raises ValueError: if ``layer`` is not ``back`` or ``fore``
    :raises ValueError: if ``char`` is not 1 character long"""
    if layer not in ['back', 'fore']:
        raise ValueError("'layer' is not 'back' or 'fore'")
    if len(char) != 1:
        raise ValueError("'char' is not 1 character long")
    i = Image.open(fp)
    pmap = i.load()
    result = {}
    ox, oy = _find_origin(i.size, anchor)
    for x in range(i.size[0]):
        for y in range(i.size[1]):
            # pmap[x, y]
            result[x-ox+2, y-oy][layer] = pmap[x, y]
            result[x-ox+2, y-oy]['char'] = char
    return result


from setuptools import setup
import tegen

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
  name = 'tegen',
  packages = ['tegen'],
  version = tegen.__version__+"",
  license ='gpl-3.0',
  description = 'Terminal game engine for Python',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = '7d',
  author_email = 'i.third.7d@protonmail.com',
  url = 'https://github.com/iiiii7d/tegen',
  download_url = f'https://github.com/iiiii7d/tegen/archive/refs/tags/v{tegen.__version__}.tar.gz',
  keywords = ['tegen', 'terminal', 'game', 'engine', 'terminal game engine', 'game development'],
  python_requires='>=3.6, <3.10',
  package_data={
    'tegen': ['examples/*'],
  },
  install_requires=[
    "blessed",
    "wcwidth",
    "pillow"
  ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: End Users/Desktop',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Games/Entertainment",
  ],
)

#commands for upload in case i forget
#python setup.py sdist
#python setup.py bdist_wheel
#twine upload dist/*